﻿--[[
	Project: KTE2 - KnowThineEnemy 2
	Author: Perm (Perm@Brutmutter - robert.meinkes@web.de)
	Original Author: NerfedWar (http://www.nerfedwar.net - nerfed.war@gmail.com)

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
]]

-------------------------------------------------------------------------------
-- Global and Static
-------------------------------------------------------------------------------

-- The KnowThineEnemy class
-- @class
KTE2 = class()

-- Global Preferences
-- @field
-- @global
KTE_Prefs = {}



-- Returns the single KTE object instance
-- @global
-- @return the KTE instance
function KTE2.getInstance() return KTE2.instance end

-- Updates the user interface if it has not been updated recently
function KTE2.updateUI()

	-- obtain a reference the KTE class instance
	local self = KTE2.getInstance()


	-- we only wan9t this running a maximum of once every 1/2 of a second
	-- mjb changed to 1 sec
	local now = Inspect.Time.Real()
	if (now < self.instance.lastUpdateUITime+0.5) then return end
	self.instance.lastUpdateUITime = now

	-- get targetname (if there is one)
	local target = Inspect.Unit.Detail('player.target')
	local player = Inspect.Unit.Detail('player')

	-- do we have any delayed data to insert
	if player then
		for _, abilityName in pairs(self.delayedPlayerAbilitiesToAdd) do
			self.UpdatePlayersDB({["abilityName"]=abilityName, ["casterName"]=player.name, ["caster"]=player.id})
		end
		self.delayedPlayerAbilitiesToAdd = {}
	end

	-- get textures
	local _,tex1 = self.kteIcon1:GetTexture()
	local _,tex2 = self.kteIcon2:GetTexture()
	local _,tex3 = self.kteIcon3:GetTexture()

	if not target or not self.playersDB[target.name] then

		-- update target textual build KTE2.SoulAbbr
		self.targetBuildFrame:SetValue()

		-- no target, clear icons
		if tex1 ~= "Icons/0.png" then self.kteIcon1:SetTexture("KTE2", "Icons/0.png") end
		if tex2 ~= "Icons/0.png" then self.kteIcon2:SetTexture("KTE2", "Icons/0.png") end
		if tex3 ~= "Icons/0.png" then self.kteIcon3:SetTexture("KTE2", "Icons/0.png") end

		self.soul1TT=''
		self.soul2TT=''
		self.soul3TT=''

		self.kteText1:SetText('')
		self.kteText2:SetText('')
		self.kteText3:SetText('')

	else
		local p1 = self.playersDB[target.name][1].Points
		local p2 = self.playersDB[target.name][2].Points
		local p3 = self.playersDB[target.name][3].Points

		p1 = tonumber(p1)
		p2 = tonumber(p2)
		p3 = tonumber(p3)

		local tmp = {}
		if p1 >= p2 and p2 >= p3 then
			tmp[1] = self.playersDB[target.name][1]
			tmp[2] = self.playersDB[target.name][2]
			tmp[3] = self.playersDB[target.name][3]
		elseif p1 >= p3 and p3 >= p2 then
			tmp[1] = self.playersDB[target.name][1]
			tmp[2] = self.playersDB[target.name][3]
			tmp[3] = self.playersDB[target.name][2]
		elseif p2 >= p1 and p1 >= p3 then
			tmp[1] = self.playersDB[target.name][2]
			tmp[2] = self.playersDB[target.name][1]
			tmp[3] = self.playersDB[target.name][3]
		elseif p2 >= p3 and p3 >= p1 then
			tmp[1] = self.playersDB[target.name][2]
			tmp[2] = self.playersDB[target.name][3]
			tmp[3] = self.playersDB[target.name][1]
		elseif p3 >= p1 and p1 >= p2 then
			tmp[1] = self.playersDB[target.name][3]
			tmp[2] = self.playersDB[target.name][1]
			tmp[3] = self.playersDB[target.name][2]
		elseif p3 >= p2 and p2 >= p1 then
			tmp[1] = self.playersDB[target.name][3]
			tmp[2] = self.playersDB[target.name][2]
			tmp[3] = self.playersDB[target.name][1]
		end

		self.playersDB[target.name] = tmp

		-- update target textual build KTE2.SoulAbbr
		self.targetBuildFrame:SetValue(target.calling, KTE2.SoulAbbr[self.playersDB[target.name][1].SoulId], self.playersDB[target.name][1].Points, KTE2.SoulAbbr[self.playersDB[target.name][2].SoulId], self.playersDB[target.name][2].Points, KTE2.SoulAbbr[self.playersDB[target.name][3].SoulId], self.playersDB[target.name][3].Points)

		-- has target, update icons
		if tex1 ~= "Icons/"..self.playersDB[target.name][1].SoulId..".png" then self.kteIcon1:SetTexture("KTE2", "Icons/"..self.playersDB[target.name][1].SoulId..".png") end
		if tex2 ~= "Icons/"..self.playersDB[target.name][2].SoulId..".png" then self.kteIcon2:SetTexture("KTE2", "Icons/"..self.playersDB[target.name][2].SoulId..".png") end
		if tex3 ~= "Icons/"..self.playersDB[target.name][3].SoulId..".png" then self.kteIcon3:SetTexture("KTE2", "Icons/"..self.playersDB[target.name][3].SoulId..".png") end

		self.soul1TT=self.playersDB[target.name][1].SoulName
		self.soul2TT=self.playersDB[target.name][2].SoulName
		self.soul3TT=self.playersDB[target.name][3].SoulName

		if KTE2.SoulAbbr[self.playersDB[target.name][1].SoulId] then self.kteText1:SetText(tostring(p1)..KTE2.SoulAbbr[self.playersDB[target.name][1].SoulId]) else self.kteText1:SetText('') end
		if KTE2.SoulAbbr[self.playersDB[target.name][2].SoulId] then self.kteText2:SetText(tostring(p2)..KTE2.SoulAbbr[self.playersDB[target.name][2].SoulId]) else self.kteText2:SetText('') end
		if KTE2.SoulAbbr[self.playersDB[target.name][3].SoulId] then self.kteText3:SetText(tostring(p3)..KTE2.SoulAbbr[self.playersDB[target.name][3].SoulId]) else self.kteText3:SetText('') end
	end
end

-- Is called when the player gains an ability
-- @param abilities table of abilities
-- @global
function KTE2.parseAbilityAddEvent(abilities)

	-- obtain a reference the KTE class instance
	local self = KTE2.getInstance()

	local player = Inspect.Unit.Detail('player')

	-- attempt to skip nil records
	if not abilities then

		for abilityId, bool in pairs(abilities) do
			-- frequent err here with nil ability detail
			if abilityId then
				local ability = Inspect.Ability.Detail(abilityId)
			end
			if ability.name and player then
				self.UpdatePlayersDB({["abilityName"]=ability.name, ["casterName"]=player.name, ["caster"]=player.id})
			elseif ability.name then
				table.insert(self.delayedPlayerAbilitiesToAdd, ability.name)
			end
		end
	end

	-- update the UI
	self.updateUI()
end


-- Is called when a unit target changes
-- By default the unit
-- @param unitSpecifier the string identifier for the unit (player.target etc...)
-- @param unitId the unique id of the unit
-- @global
function KTE2.parseTargetChangedEvent(evt)

	-- obtain a reference the KTE class instance
	local self = KTE2.getInstance()

	-- get current time
	local now = Inspect.Time.Real() -- API CALL

	-- gets set to true if we need to update the UI after
	local playerTargetEventIncluded = false

	-- loop through event and retrieve unitId and unitSpecifier for each entry
	for unitId, unitSpecifier in pairs(evt) do

		-- check we have a valid unit specifier
		if unitSpecifier and self.validUnitSpecifiers[unitSpecifier] then

			-- we only update the UI if we have a valid update unit specifier
			if unitSpecifier and self.validUpdateUnitSpecifiers[unitSpecifier] then playerTargetEventIncluded = true end

			-- we only continue if the unitId for the unitSpecifier is different from last time
			if self.lastUnitSpecifierToUnitId[unitSpecifier] ~= unitId then

				-- update last unit specifier
				self.lastUnitSpecifierToUnitId[unitSpecifier] = unitId

				-- only scan for buffs if we haven't done so in within the last unitSpecifierLastBuffScanFrequency seconds
				if not self.unitSpecifierLastBuffScan[unitSpecifier] then self.unitSpecifierLastBuffScan[unitSpecifier] = 0 end
				if (now > self.unitSpecifierLastBuffScan[unitSpecifier]+self.unitSpecifierLastBuffScanFrequency) then
					self.unitSpecifierLastBuffScan[unitSpecifier] = now

					-- perform buff scan
					-- performance notes: nI've got to call Inspect.Buff.List
					-- to get the initial buff listing, then I've got to call
					--Inspect.Buff.Detail to get the 'actual' buff info to find
					-- out the source of the ability. I COULD keep a cache of unitID
					-- to unitNames so I wouldnb't have to call Inspect.Unit.Detail!
					local buffList = Inspect.Buff.List(unitId) -- API CALL
					if buffList then
						for j, k in pairs(buffList) do

							local buffDetail = Inspect.Buff.Detail(unitId, j) -- API CALL
							local caster = Inspect.Unit.Detail(buffDetail.caster) -- API CALL (could be replaced with a cache lockup!)
							if caster then
								-- update the db
								self.UpdatePlayersDB({["abilityName"]=buffDetail.name, ["casterName"]=caster.name, ["caster"]=buffDetail.caster})
							end
						end
					end
				end
			end
		end
	end

	-- update UI
	if playerTargetEventIncluded then self.updateUI() end
end


-- Called when something in the unit tree fires a castbar event
-- @param evt the castbar event
function KTE2.parseCastbarEvent(evt)

	-- obtain a reference the KTE class instance
	local self = KTE2.getInstance()

	-- gets set to true if we need to update the UI after
	local playerTargetEventIncluded = false

	local unitCastBarDetails = Inspect.Unit.Castbar(evt) -- API CALL

	for unitId, abilityTable in pairs(unitCastBarDetails) do

		-- check we have a valid unit specifier
		local unitSpecifier = Inspect.Unit.Lookup(unitId) -- API CALL
		if unitSpecifier and self.validUnitSpecifiers[unitSpecifier] then

			-- we only update the UI if we have a valid update unit specifier
			if unitSpecifier and self.validUpdateUnitSpecifiers[unitSpecifier] then playerTargetEventIncluded = true end

			-- update the db
			local caster = Inspect.Unit.Detail(unitId) -- API CALL
			self.UpdatePlayersDB({["abilityName"]=abilityTable.abilityName, ["casterName"]=caster.name, ["caster"]=unitId})
		end
	end

	-- update UI
	if playerTargetEventIncluded then self.updateUI() end
end

-- Called when someone receives a buff
-- @param unit the unitid who gained the buff (note: we don't really care, we're more interested in who cast the buff!)
-- @param buffs an array of buff ids gained
function KTE2.parseBuffAddEvent(unit, buffs)

	-- get reference to KTE instance
	local self = KTE2.getInstance()

	-- gets set to true if we need to update the UI after
	local playerTargetEventIncluded = false

	-- check we have a valid unit specifier
	-- we're not interested in buffs/debuffs on anyone outside out unit specifier list,
	-- too CPU intensive expecially when AoEing in dungeons!
	local unitSpecifier = Inspect.Unit.Lookup(unit)  -- API CALL
	if unitSpecifier and self.validUnitSpecifiers[unitSpecifier] then

		-- we only update the UI if we have a valid update unit specifier
		if unitSpecifier and self.validUpdateUnitSpecifiers[unitSpecifier] then playerTargetEventIncluded = true end

		local buffsList = Inspect.Buff.Detail(unit, buffs) -- API CALL

		for j, buffDetail in pairs(buffsList) do

			-- 1. Store the id and name of the buff in our buffDB
			self.buffDB[buffDetail.id]=buffDetail.name

			-- 2. Update the soul record of the unit, but only if they were source of the buff onto themselves
			local buffer = Inspect.Unit.Detail(buffDetail.caster)  -- API CALL
			if buffer and buffer.id == unit then
				self.UpdatePlayersDB({["abilityName"]=buffDetail.name, ["casterName"]=buffer.name, ["caster"]=buffer.id})
			end
		end

		-- update UI
		if playerTargetEventIncluded then self.updateUI() end

	end
end


-- Called when someone loses a buff, we only check this event to determine if a reflect ability has been removed, nothing more.
-- @param unit the unitid who lost the buff
-- @param buffs an array of buff ids lost
function KTE2.parseBuffRemoveEvent(unit, buffs)

	-- get reference to KTE instance
	local self = KTE2.getInstance()

	-- gets set to true if we need to update the UI after
	local playerTargetEventIncluded = false

	-- obtain the player record (or create one if one does not exist)
	local buffedUnit = Inspect.Unit.Detail(unit) -- API CALL

	-- check we have a valid unit specifier
	-- we're not interested in buffs/debuffs on anyone outside out unit specifier list,
	local unitSpecifier = Inspect.Unit.Lookup(unit) -- API CALL
	if unitSpecifier and self.validUnitSpecifiers[unitSpecifier] then

		local playerRecord = self:getPlayerRecord(buffedUnit.name)

		-- check if the buff removed was one of the one-off reflect buffs, if so we need to add a flag so that
		-- the next ability by the unit that lost the buff is ignored.
		for j, k in pairs(buffs) do
			if self.buffDB[j] then
				local buffName = self.buffDB[j]
				self.buffDB[j] = nil -- clean out buff
				if buffName=="Reflective Command" or buffName=="Glassy Reflection" or buffName=="Reversal of Fortune" then

					self:debugMsg("Reflect buff "..buffName.." removed, ignoring next ability(ies)...")

					-- we only update the UI if we have a valid update unit specifier
					if unitSpecifier and self.validUpdateUnitSpecifiers[unitSpecifier] then playerTargetEventIncluded = true end

					playerRecord.ignore = true
				end
			else
				-- lets be safe and ignore the next ability if its missing from our buffDB cache...

				-- we only update the UI if we have a valid update unit specifier
				if unitSpecifier and self.validUpdateUnitSpecifiers[unitSpecifier] then playerTargetEventIncluded = true end

				playerRecord.ignore = true
			end
		end

		-- update UI
		if playerTargetEventIncluded then self.updateUI() end
	end
end


-- Called when there is a damage or healing event
-- @param evt the combat heal/damage event
function KTE2.parseCombatEvent(evt)

	-- get reference to KTE instance
	local self = KTE2.getInstance()

	-- gets set to true if we need to update the UI after
	local playerTargetEventIncluded = false

	-- check we have a valid unit specifier
	local unitSpecifier = Inspect.Unit.Lookup(evt.caster) -- API CALL
	if unitSpecifier and self.validUnitSpecifiers[unitSpecifier] then

		-- we only update the UI if we have a valid update unit specifier
		if unitSpecifier and self.validUpdateUnitSpecifiers[unitSpecifier] then playerTargetEventIncluded = true end

		-- get key params
		local abilityName = evt.abilityName
		local casterName = evt.casterName
		local caster = evt.caster

		-- return if caster name not available
		if not casterName then return end

		-- get player record
		local playerRecord = self:getPlayerRecord(casterName)

		-- if this ability is on the ignore list then we can ignore it :)
		if abilityName then
			if self.combatEventAbilityIgnoreList[abilityName] then return end
		end

		if abilityName then
			if casterName then
				if self.abilityDB[casterName] then
					  if self.abilityDB[casterName][abilityName] then return end
				end
			end
		end

		-- if we already know the caster has this ability we can stop right here!
		if not self.abilityDB[casterName] then self.abilityDB[casterName] = {} end

		-- loop through buffs on caster (not such an issue anymore as we only do this for new abilities for the caster :)
		local buffList = Inspect.Buff.List(caster) -- API CALL
		if buffList then
			for j, k in pairs(buffList) do
				local buffDetail = Inspect.Buff.Detail(caster, j) -- API CALL

				-- 1. does the player have a duration based reflect buff active?
				-- is so the ability could be a reflect!
				if buffDetail.name=="Reflective Command" or buffDetail.name=="Glassy Reflection" or buffDetail.name=="Reversal of Fortune" then
					self:debugMsg("Ignoring "..abilityName.." due to active reflect buff "..buffDetail.name..".")
					return
				end

				-- 2. does the player have a buff equal to the name of the damage ability and is the buff by someone else?
				-- should resolve issues with abilities like spotters where the caster gains a buff that does ability damage and doesn't own the actual ability itself.
				if Inspect.Unit.Detail(buffDetail.caster) then
					local bufferName = Inspect.Unit.Detail(buffDetail.caster).name
					if bufferName ~= casterName and buffDetail.name == abilityName then
						return
					elseif buffDetail.name == abilityName and playerRecord.ignore then
						self:debugMsg("Ignoring "..abilityName.." due to reflect buff "..buffDetail.name..".")
						return
					end
				end
			end
		end

		-- if the player has the ignore flag then quit right here as we're probably dealing with a reflect
		if not playerRecord then return end
		if playerRecord.ignore then
			self:debugMsg("Ignoring "..abilityName.." due to recent reflect buff.")
			playerRecord.ignore=false
			return
		end

		-- do the db update
		self.UpdatePlayersDB(evt)

		-- update UI
		if playerTargetEventIncluded then self.updateUI() end
	end
end


-- Updates the players database
-- @param evt Event data
function KTE2.UpdatePlayersDB(evt)

	-- get reference to KTE instance
	local self = KTE2.getInstance()

	-- make sure the souls database is loaded
	if not KTE2.SoulsDB then return end

	-- get key params
	local abilityName = evt.abilityName
	local casterName = evt.casterName
	local caster = evt.caster

	-- we ignore if we don't have the the minimum number of parameters
	if not abilityName or not casterName or not KTE2.SoulsDB[abilityName] then return end

	-- retrieve Soul and Points for the ability
	local soulid = KTE2.SoulsDB[abilityName].SoulId
	local soul = KTE2.SoulsDB[abilityName].Soul
	local points = KTE2.SoulsDB[abilityName].Points

	-- if we dont get a match then quit
	if not soul or not points then return end

	-- get player record
	local playerRecord = self:getPlayerRecord(casterName)

	-- copy existing souls
	local dbEntryCopy = self.playersDB[casterName]

	-- get current time
	local now = Inspect.Time.Real()

	if soul == self.playersDB[casterName][1].SoulName then
		dbEntryCopy[1].SoulId = soulid
		dbEntryCopy[1].SoulName = soul
		dbEntryCopy[1].LastUpdate = now
		dbEntryCopy[1].Points = math.max(self.playersDB[casterName][1].Points,points)
	elseif soul == self.playersDB[casterName][2].SoulName then
		dbEntryCopy[2].SoulId = soulid
		dbEntryCopy[2].SoulName = soul
		dbEntryCopy[2].LastUpdate = now
		dbEntryCopy[2].Points = math.max(self.playersDB[casterName][2].Points,points)
	elseif soul == self.playersDB[casterName][3].SoulName then
		dbEntryCopy[3].SoulId = soulid
		dbEntryCopy[3].SoulName = soul
		dbEntryCopy[3].LastUpdate = now
		dbEntryCopy[3].Points = math.max(self.playersDB[casterName][3].Points,points)
	else

		-- find empty slot
		if self.playersDB[casterName][1].SoulName == '' then
			dbEntryCopy[1].SoulId = soulid
			dbEntryCopy[1].SoulName = soul
			dbEntryCopy[1].LastUpdate = now
			dbEntryCopy[1].Points = math.max(self.playersDB[casterName][1].Points,points)
		elseif self.playersDB[casterName][2].SoulName == '' then
			dbEntryCopy[2].SoulId = soulid
			dbEntryCopy[2].SoulName = soul
			dbEntryCopy[2].LastUpdate = now
			dbEntryCopy[2].Points = math.max(self.playersDB[casterName][2].Points,points)
		elseif self.playersDB[casterName][3].SoulName == '' then
			dbEntryCopy[3].SoulId = soulid
			dbEntryCopy[3].SoulName = soul
			dbEntryCopy[3].LastUpdate = now
			dbEntryCopy[3].Points = math.max(self.playersDB[casterName][3].Points,points)
		else
			-- we have a different soul! = different build! = clean out and start again!
			dbEntryCopy = {}
			dbEntryCopy[1] = {["SoulId"]=soulid,["SoulName"]=soul,["Points"]=points,["LastUpdate"]=now}
			dbEntryCopy[2] = {["SoulId"]=0,["SoulName"]='',["Points"]=0,["LastUpdate"]=0}
			dbEntryCopy[3] = {["SoulId"]=0,["SoulName"]='',["Points"]=0,["LastUpdate"]=0}
			self.abilityDB[casterName] = {}
		end

	end

	-- update abilitydb
	if not self.abilityDB[casterName] then self.abilityDB[casterName] = {} end
	self.abilityDB[casterName][abilityName] = true

	-- update souls db
	self.playersDB[casterName] = dbEntryCopy
end


-------------------------------------------------------------------------------
-- Class Members
-------------------------------------------------------------------------------

-- KTE class constructore that loads preferences, creates the UI and initiates event handlers.
function KTE2:init()

	-- store reference to self, used by KTE2.getInstance() to return the single KTE instance
	-- @field
	KTE2.instance = self

	-- initialise database for players and mobs (gets reset each login/reloadui)
	-- @field
	self.playersDB = {}

	-- initialise database for buff.id to buff.name mapping (we need this when a buff gets removed as we can no longer look it up)
	-- @field
	self.buffDB = {}

	-- initialise database for who has what abilities so we can ignore future casts of an ability
	-- @field
	self.abilityDB = {}

	-- initialise a table that holds any abilities to add to your playerDB during the next scheduled update
	-- we need this in the cases when for some reason we can't add the records at the time of the event
	-- possibly due to the player object not being initialised at that time.
	self.delayedPlayerAbilitiesToAdd = {}

	-- stores when the updateUI function was last called
	-- @field
	self.lastUpdateUITime = Inspect.Time.Real()

	-- stores when the addon last scanned a unit for their buff list
	-- @field
	self.unitSpecifierLastBuffScan = {}

	-- max frequency in seconds of buff scans
	-- @field
	self.unitSpecifierLastBuffScanFrequency = 1

	-- list of valid unit spcifiers
	-- @field
	self.validUnitSpecifiers = {
		['player'] = true, ['player.target'] = true,
		['group01'] = true, ['group01.target'] = true,
		['group02'] = true, ['group02.target'] = true,
		['group03'] = true, ['group03.target'] = true,
		['group04'] = true, ['group04.target'] = true,
		['group05'] = true, ['group05.target'] = true,
		['group06'] = true, ['group06.target'] = true,
		['group07'] = true, ['group07.target'] = true,
		['group08'] = true, ['group08.target'] = true,
		['group09'] = true, ['group09.target'] = true,
		['group10'] = true, ['group10.target'] = true,
		['group11'] = true, ['group11.target'] = true,
		['group12'] = true, ['group12.target'] = true,
		['group13'] = true, ['group13.target'] = true,
		['group14'] = true, ['group14.target'] = true,
		['group15'] = true, ['group15.target'] = true,
		['group16'] = true, ['group16.target'] = true,
		['group17'] = true, ['group17.target'] = true,
		['group18'] = true, ['group18.target'] = true,
		['group19'] = true, ['group19.target'] = true,
		['group20'] = true, ['group01.target'] = true
	}

	-- list of valid unit spcifiers that will trigger update
	-- @field
	self.validUpdateUnitSpecifiers = {
		['player'] = true, ['player.target'] = true
	}

	-- list of abilities that are just to problematic to deal with when they heal/damage
	-- only abilities with cast times are placed here as we'll pick those abilities up
	-- via the castbar event anyway :)
	-- @field
	self.combatEventAbilityIgnoreList = {
		['Soul Drain'] = true,
		['other'] = true
	}

	-- stores last unitSpecifier to unitId mapping
	-- @field
	self.lastUnitSpecifierToUnitId = {}

	-- create the context
	-- @field
	self.context = UI.CreateContext("KTEContext")

	-- whether to show debug messages or not
	self.debug = false

	-- make sure we have a config store for the bookPosition x,y
	if not KTE_Prefs then KTE_Prefs = {} end
	if not KTE_Prefs['x'] then KTE_Prefs['x'] = 0 end
	if not KTE_Prefs['y'] then KTE_Prefs['y'] = 0 end

	-- init tooltip storage
	-- @field
	self.soul1TT=''
	self.soul2TT=''
	self.soul3TT=''

	-- create the draggable ui
	-- @field
	self.leftDown = false
	self.originalXDiff = 0
	self.originalYDiff = 0
	self:createUI()

	-- initiate event handlers
	table.insert(Event.Combat.Damage, { KTE2.parseCombatEvent, "KTE2", "KTE2.parseCombatEvent" })
	table.insert(Event.Combat.Heal, { KTE2.parseCombatEvent, "KTE2", "KTE2.parseCombatEvent" })
	table.insert(Event.Unit.Castbar, { KTE2.parseCastbarEvent, "KTE2", "KTE2.parseCastbarEvent" })
	table.insert(Event.Buff.Add, {KTE2.parseBuffAddEvent, "KTE2", "KTE2.parseBuffAddEvent"})
	table.insert(Event.Buff.Remove, {KTE2.parseBuffRemoveEvent, "KTE2", "KTE2.parseBuffRemoveEvent"})
	table.insert(Event.Unit.Add, { KTE2.parseTargetChangedEvent, "KTE2", "KTE2.parseTargetChangedEvent" })
	table.insert(Event.Unit.Remove, { KTE2.parseTargetChangedEvent, "KTE2", "KTE2.parseTargetChangedEvent" })
	table.insert(Event.Ability.New.Add, { KTE2.parseAbilityAddEvent, "KTE2", "KTE2.parseAbilityAddEvent" })
	table.insert(Event.System.Update.Begin, { KTE2.updateUI, "KTE2", "KTE2.updateUI" })

	-- output loaded message
	print("KTE2 v0.5 loaded... Have fun!")
end


-- Output a debug message if debug is enabled
-- @see enableDebug
-- @param msg The debug message
function KTE2:debugMsg(msg)
	if(self.debug) then
		print("[DEBUG] "..msg)
	end
end


-- creates a blank player database record or returns an existing one
-- @param the name of the player or mob
-- @return the player object
function KTE2:getPlayerRecord(name)
	if not name then return nil end
	if not self.playersDB[name] then
		self.playersDB[name] = {}
		self.playersDB[name][1] = {["SoulId"]=0,["SoulName"]='',["Points"]=0,["LastUpdate"]=0}
		self.playersDB[name][2] = {["SoulId"]=0,["SoulName"]='',["Points"]=0,["LastUpdate"]=0}
		self.playersDB[name][3] = {["SoulId"]=0,["SoulName"]='',["Points"]=0,["LastUpdate"]=0}
	end
	return self.playersDB[name]
end


-- creates the HUD
function KTE2:createUI()
	-- create the parent frame and make it draggable
	self.kteFrame = UI.CreateFrame("Texture", "KTEFrame", self.context)
	self.kteFrame:SetTexture("KTE2", "Icons/frame.png")

	-- self.kteFrame:ResizeToTexture()
	self.kteFrame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", KTE_Prefs['x'], KTE_Prefs['y'])
	self.kteFrame:SetLayer(1)
	self.kteFrame:SetVisible(true)
	function self.kteFrame.Event:LeftDown()
		self.leftDown = true
		local mouse = Inspect.Mouse()
		self.originalXDiff = mouse.x - self:GetLeft()
		self.originalYDiff = mouse.y - self:GetTop()
	end
	function self.kteFrame.Event:LeftUp()
		self.leftDown = false
	end
	function self.kteFrame.Event:LeftUpoutside()
		self.leftDown = false
	end
	function self.kteFrame.Event:MouseMove(x, y)
		if not self.leftDown then
			return
		end
		self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x - self.originalXDiff, y - self.originalYDiff)
		KTE_Prefs['x'] = x - self.originalXDiff
		KTE_Prefs['y'] = y - self.originalYDiff
	end

	self.kteTT = UI.CreateFrame("Text", "KTETT", self.context)
	self.kteTT:SetText("")
	self.kteTT:SetFontSize(12)
	self.kteTT:SetPoint("CENTER", self.kteFrame, "CENTER", 0, -50)
	self.kteTT:SetLayer(10)
	self.kteTT:SetVisible(true)

	-- create the 3 'sub' role texture frames
	self.kteIcon1 = UI.CreateFrame("Texture", "KTEIcon1", self.context)
	self.kteIcon1:SetTexture("KTE2", "Icons/0.png")
	-- self.kteIcon1:ResizeToTexture()
	self.kteIcon1:SetPoint("TOPLEFT", self.kteFrame, "TOPLEFT", 1, 1)
	self.kteIcon1:SetLayer(10)
	self.kteIcon1:SetVisible(true)
	self.kteIcon1:SetMouseMasking("limited")
	function self.kteIcon1.Event:MouseIn()
		KTE2.instance.kteTT:SetText(KTE2.instance.soul1TT)
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end
	function self.kteIcon1.Event:MouseMove()
		KTE2.instance.kteTT:SetText(KTE2.instance.soul1TT)
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end
	function self.kteIcon1.Event:MouseOut()
		KTE2.instance.kteTT:SetText('')
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end

	self.kteIcon2 = UI.CreateFrame("Texture", "KTEIcon2", self.context)
	self.kteIcon2:SetTexture("KTE2", "Icons/0.png")
	-- self.kteIcon2:ResizeToTexture()
	self.kteIcon2:SetPoint("TOPLEFT", self.kteFrame, "TOPLEFT", 37, 1)
	self.kteIcon2:SetLayer(10)
	self.kteIcon2:SetVisible(true)
	self.kteIcon2:SetMouseMasking("limited")
	function self.kteIcon2.Event:MouseIn()
		KTE2.instance.kteTT:SetText(KTE2.instance.soul2TT)
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end
	function self.kteIcon2.Event:MouseMove()
		KTE2.instance.kteTT:SetText(KTE2.instance.soul2TT)
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end
	function self.kteIcon2.Event:MouseOut()
		KTE2.instance.kteTT:SetText('')
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end

	self.kteIcon3 = UI.CreateFrame("Texture", "KTEIcon3", self.context)
	self.kteIcon3:SetTexture("KTE2", "Icons/0.png")
	-- self.kteIcon3:ResizeToTexture()
	self.kteIcon3:SetAlpha(0.9)
	self.kteIcon3:SetPoint("TOPLEFT", self.kteFrame, "TOPLEFT", 72, 1)
	self.kteIcon3:SetLayer(10)
	self.kteIcon3:SetVisible(true)
	self.kteIcon3:SetMouseMasking("limited")
	function self.kteIcon3.Event:MouseIn()
		KTE2.instance.kteTT:SetText(KTE2.instance.soul3TT)
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end
	function self.kteIcon3.Event:MouseMove()
		KTE2.instance.kteTT:SetText(KTE2.instance.soul3TT)
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end
	function self.kteIcon3.Event:MouseOut()
		KTE2.instance.kteTT:SetText('')
		KTE2.instance.kteTT:SetPoint("CENTER", KTE2.instance.kteFrame, "CENTER", 0, -50)
	end

	-- create the 3 'sub' role points text frames
	self.kteText1 = UI.CreateFrame("Text", "KTEText1", self.context)
	self.kteText1:SetText("")
	self.kteText1:SetFontSize(12)
	self.kteText1:SetPoint("CENTER", self.kteIcon1, "CENTER", 0, 27)
	self.kteText1:SetLayer(10)
	self.kteText1:SetVisible(true)

	self.kteText2 = UI.CreateFrame("Text", "KTEText2", self.context)
	self.kteText2:SetText("")
	self.kteText2:SetFontSize(12)
	self.kteText2:SetPoint("CENTER", self.kteIcon2, "CENTER", 0, 27)
	self.kteText2:SetLayer(10)
	self.kteText2:SetVisible(true)

	self.kteText3 = UI.CreateFrame("Text", "KTEText3", self.context)
	self.kteText3:SetText("")
	self.kteText3:SetFontSize(12)
	self.kteText3:SetPoint("CENTER", self.kteIcon3, "CENTER",0, 27)
	self.kteText3:SetLayer(10)
	self.kteText3:SetVisible(true)

	-- text based ui elements below here
	self.targetBuildFrame = KTE2:createBuildFrame("TargetBuild", self.context, UI.Native.PortraitTarget, -40, 130, "TOPRIGHT", "TOPRIGHT")
end


function KTE2:createBuildFrame(name, context, parent, x, y, p1, p2)
	local aFrame = UI.CreateFrame("Text", name, context)
	aFrame:SetFontSize(16)
	aFrame:SetFontColor(8, 8,8, 8)
	aFrame:SetPoint(p1, parent, p2, x, y)
	aFrame:SetLayer(99)
	aFrame:SetVisible(true)

	function aFrame:SetValue(class, soul1Abbrev, soul1Points, soul2Abbrev, soul2Points, soul3Abbrev, soul3Points)
		if not class then
			aFrame:SetText("")
		else
			aFrame:SetText(KTE2.ClassDB[class])
		end
	end
	return aFrame
end


-- Returns the build of a player
-- Can be called externally: KTW.getInstance():getRoles(name)
-- @param name Name of a player or mob.
function KTE2:getSouls(name)
	if self.playersDB[name] then
		return self.playersDB[names]
	end
	return false
end


-- Create KTE instance
-- @event initialisation after addon initialisation
table.insert(Event.Addon.Load.End, {
	function(addon) if addon == "KTE2" then KTE2() end
	end, "KTE2", "Initialise"
})


